"""Sphinx conf."""

import sys
import tomllib
from datetime import date
from os import path

sys.path.append(path.abspath(".."))

with open("../pyproject.toml", "rb") as fh:
    toml = tomllib.load(fh)

extensions = [
    "sphinx.ext.coverage",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "myst_parser",
]

pyproject = toml["project"]

project = pyproject["name"]
version = pyproject["version"]
release = pyproject["version"]

copyright = f"{date.today().year:d} Erik Brinkman"  # noqa: A001
author = "Erik Brinkman"



"""Tests for the homeomorphism builder."""

import numpy as np

import homeotopy


def test_composition() -> None:
    """Test composing transforms with homeomorphism."""
    rng = np.random.default_rng()
    homeo = homeotopy.homeomorphism(homeotopy.simplex(), homeotopy.sphere())

    simplex = rng.dirichlet(np.ones(4), (2, 5))
    sphere = homeo(simplex)
    assert np.allclose(np.linalg.norm(sphere, 2, -1), 1)

    inv = ~homeo
    actual = inv(sphere)
    assert np.allclose(simplex, actual)



import math
from dataclasses import dataclass

import numpy as np
from numpy.typing import NDArray

from ._homeomorphism import Topology


@dataclass(frozen=True, slots=True)
class Ball(Topology):
    """The topology of the p-norm ball.

    This represents all points in R^n s.t. ||x||_p < 1, although it should also
    work fo the boundary.
    """

    p: float

    def __post_init__(self) -> None:
        if self.p <= 0:
            raise ValueError(f"p must be greater than or equal to 0: {self.p:g}")

    def to_inf_ball(self, points: NDArray[np.float64]) -> NDArray[np.float64]:
        if self.p == math.inf:
            return points
        else:
            tiny = np.finfo(points.dtype).smallest_normal
            source = np.linalg.norm(points, self.p, -1)
            target = np.abs(points).max(-1) + tiny
            return np.clip(points * (source / target)[..., None], -1, 1)  # type: ignore

    def from_inf_ball(self, points: NDArray[np.float64]) -> NDArray[np.float64]:
        if self.p == math.inf:
            return points
        else:
            tiny = np.finfo(points.dtype).smallest_normal
            source = np.abs(points).max(-1)
            target = np.linalg.norm(points, self.p, -1) + tiny
            return points * (source / target)[..., None]  # type: ignore


def ball(p: float = 2.0) -> Ball:
    """Create a topology of the interior of the p-norm unit ball."""
    return Ball(p)



"""Tests for sphere projection."""

import numpy as np

import homeotopy


def test_known() -> None:
    """Test projecting some known points."""
    sphere = homeotopy.sphere()

    sphere_points = np.array([[1, 0, 0], [0, 1, 0], [1 / 2, 0, 3**0.5 / 2]], "f8")
    inf_ball_points = np.array([[1, 1], [np.tanh(1), 0], [0, np.tanh(3**0.5)]], "f8")
    assert np.allclose(sphere.to_inf_ball(sphere_points), inf_ball_points)
    assert np.allclose(sphere.from_inf_ball(inf_ball_points), sphere_points)


def test_random() -> None:
    """Test that random projections follow invariants."""
    rng = np.random.default_rng(0)
    sphere = homeotopy.sphere()

    inf_ball_points = rng.uniform(-1, 1, (3, 4, 5))
    sphere_points = sphere.from_inf_ball(inf_ball_points)
    assert np.allclose(sphere.to_inf_ball(sphere_points), inf_ball_points)
    assert np.allclose(np.linalg.norm(sphere_points, 2, -1), 1)



"""Tests for cube projection."""

import numpy as np

import homeotopy


def test_known() -> None:
    """Test projecting known points."""
    cube = homeotopy.cube()

    cube_points = np.array([[0, 0, 0], [1, 1, 1], [0.5, 0.5, 0.5]], "f8")
    inf_ball_points = np.array([[-1, -1, -1], [1, 1, 1], [0, 0, 0]], "f8")
    assert np.allclose(cube.to_inf_ball(cube_points), inf_ball_points)
    assert np.allclose(cube.from_inf_ball(inf_ball_points), cube_points)


def test_random() -> None:
    """Test projecting random points satisfy invariants."""
    rng = np.random.default_rng(0)
    cube = homeotopy.cube()

    inf_ball_points = rng.uniform(-1, 1, (3, 4, 5))
    cube_points = cube.from_inf_ball(inf_ball_points)
    assert np.allclose(cube.to_inf_ball(cube_points), inf_ball_points)
    assert np.all(0 <= cube_points) and np.all(cube_points <= 1)



"""Tests for projection to p-ball."""

import numpy as np
import pytest

import homeotopy


def test_known() -> None:
    """Test projecting known points."""
    ball = homeotopy.ball()

    ball_points = np.array(
        [
            [0, 0, 0],
            [0, 0, 1],
            [0, 1 / 2**0.5, 1 / 2**0.5],
            [1 / 3**0.5, 1 / 3**0.5, -1 / 3**0.5],
            [1 / 2, -1 / 2, 1 / 2],
        ],
        "f8",
    )
    inf_ball_points = np.array(
        [
            [0, 0, 0],
            [0, 0, 1],
            [0, 1, 1],
            [1, 1, -1],
            [3**0.5 / 2, -(3**0.5) / 2, 3**0.5 / 2],
        ],
        "f8",
    )
    assert np.allclose(ball.to_inf_ball(ball_points), inf_ball_points)
    assert np.allclose(ball.from_inf_ball(inf_ball_points), ball_points)


@pytest.mark.parametrize("p", [0.5, 1, 2, 3])
def test_random(p: float) -> None:
    """Test projecting random points satisfy invariants."""
    rng = np.random.default_rng(0)
    ball = homeotopy.ball(p)

    inf_ball_points = rng.uniform(-1, 1, (3, 4, 5))
    inf_norms = np.abs(inf_ball_points).max(-1)
    ball_points = ball.from_inf_ball(inf_ball_points)
    ball_norms = np.linalg.norm(ball_points, ball.p, -1)

    assert np.allclose(ball.to_inf_ball(ball_points), inf_ball_points)
    assert np.all(np.linalg.norm(ball_points, ball.p, -1) <= 1)
    assert np.allclose(inf_norms, ball_norms)


def test_inf() -> None:
    """Test that inf-ball is a noop."""
    rng = np.random.default_rng(0)
    ball = homeotopy.ball(np.inf)
    points = rng.uniform(-1, 1, (3, 4, 5))
    assert np.allclose(points, ball.from_inf_ball(points))
    assert np.allclose(points, ball.to_inf_ball(points))


def test_invalid_balls() -> None:
    """Exception thrown when p is invalid."""
    with pytest.raises(ValueError):
        homeotopy.ball(0)



"""A library for creating some standard homeomorphisms.

This library is based around a set of `Topologies`. Calling `homeomorphism` with
two Topologies creates a homeomorphism from one topology to the other.
Topoligies should specify their domains, but where unspecified, the topoligies
try to conform to a reasonable standard domain. The homeomorphism should work
for closed set elements too, but thos elements may not be bijective.

Remarks
-------
It's probably important to note that floating point numbers are not real
numbers, and so none of these are really bijective at all.

Also note that this library does not define homeotopies. It's just named this
have to "py" in the name.
"""

from ._ball import Ball, ball
from ._cube import Cube, cube
from ._homeomorphism import Homeomorphism, Topology, homeomorphism
from ._plane import Plane, plane
from ._simplex import Simplex, simplex
from ._sphere import Sphere, sphere

__all__ = (
    "Homeomorphism",
    "Topology",
    "homeomorphism",
    "Ball",
    "ball",
    "Cube",
    "cube",
    "Plane",
    "plane",
    "Sphere",
    "sphere",
    "Simplex",
    "simplex",
)



from dataclasses import dataclass
from functools import cache

import numpy as np
from numpy.typing import NDArray

from ._homeomorphism import Topology


@dataclass(frozen=True, slots=True)
class Plane(Topology):
    """The topology of the euclidian plane.

    This represents all points in R^n, but boundary points map to inf.

    Remarks
    -------
    While translations will keep all points valid, this will try to keep points
    at the "center" of the space mapped to (0, 0, ..., 0).
    """

    def to_inf_ball(self, points: NDArray[np.float64]) -> NDArray[np.float64]:
        info = np.finfo(points.dtype)
        # max and inf will both get promoted to 1
        clipped = np.clip(points, info.min, info.max)
        return clipped / (1 + np.abs(clipped))

    def from_inf_ball(self, points: NDArray[np.float64]) -> NDArray[np.float64]:
        clipped = np.clip(points, -1, 1)
        # this triggers at -1 and 1
        with np.errstate(divide="ignore"):
            return clipped / (1 - np.abs(clipped))


@cache
def plane() -> Plane:
    """Create a topology of the euclidian plane."""
    return Plane()



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass

import numpy as np
from numpy.typing import NDArray


class Topology(ABC):
    """An abstract topological space.

    For this library to work, each Topologiy should define a homeomorphism from
    it to the inf-norm ball.

    Remarks
    -------
    `to_inf_ball` and `from_inf_ball` are not meant to be called in isolation,
    but rather used in combination with `homeomorphism`.
    """

    @abstractmethod
    def to_inf_ball(self, points: NDArray[np.float64]) -> NDArray[np.float64]:
        """Map a set of points in this topology to the inf-ball.

        Parameters
        ----------
        points : (..., d_in)
            A set of points in the input topological space

        Returns
        -------
        points : (..., d_out)
            A set of points in the inf-norm ball, e.g. -1 < x_i < 1 for points
            in the open topology, but points can be mapped to the border for
            border points in the source topology.
        """
        ...  # pragma: no cover

    @abstractmethod
    def from_inf_ball(self, points: NDArray[np.float64]) -> NDArray[np.float64]:
        """Map a set of points from the inf-ball to this topology.

        Parameters
        ----------
        points : (..., d_in)
            A set of points in the inf-norm ball, e.g. -1 < x_i < 1 for points
            in the open topology, but points on the boarder should be handled as
            well.

        Returns
        -------
        points : (..., d_out)
            A set of points in the topological space.
        """
        ...  # pragma: no cover


@dataclass(frozen=True, slots=True)
class Homeomorphism:
    """A homeomorphism from source to target.

    Homeomorphisms can be called on points in the source domain to map them to
    points in the target domain. They can also be `inverted` to create the
    inverse mapping.


    Example
    -------

        from homeotopy import homeomorphism, ball, simplex
        import numpy as np

        forward = homeomorphism(ball(1), simplex())
        backward = ~forward

        ball_points = ...
        simplex_points = forwad(ball_points)
        backward(simplex_points)

    """

    source: Topology
    target: Topology

    def __invert__(self) -> Homeomorphism:
        return Homeomorphism(self.target, self.source)

    def __call__(self, points: NDArray[np.float64]) -> NDArray[np.float64]:
        return self.target.from_inf_ball(self.source.to_inf_ball(points))


def homeomorphism(source: Topology, target: Topology) -> Homeomorphism:
    """Create a Homeomorphism from a source and a target topology."""
    return Homeomorphism(source, target)



"""Tests for simplex projections."""

import numpy as np

import homeotopy


def test_known() -> None:
    """Test projecting known points."""
    simplex = homeotopy.simplex()

    simplex_points = np.array(
        [[1, 0, 0], [0, 1, 0], [0, 0, 1], [0, 1 / 2, 1 / 2], [1 / 3, 1 / 3, 1 / 3]],
        "f8",
    )
    inf_ball_points = np.array(
        [[-1, -1 / 3**0.5], [1, -1 / 3**0.5], [0, 1], [1, 1 / 3**0.5], [0, 0]], "f8"
    )
    assert np.allclose(simplex.to_inf_ball(simplex_points), inf_ball_points)
    assert np.allclose(simplex.from_inf_ball(inf_ball_points), simplex_points)


def test_random() -> None:
    """Test projecting random points satisfy invariants."""
    rng = np.random.default_rng(0)
    simplex = homeotopy.simplex()

    inf_ball_points = rng.uniform(-1, 1, (3, 4, 5))
    simplex_points = simplex.from_inf_ball(inf_ball_points)
    assert np.allclose(simplex.to_inf_ball(simplex_points), inf_ball_points)



from dataclasses import dataclass
from functools import cache

import numba as nb
import numpy as np
from numpy.typing import NDArray

from ._homeomorphism import Topology

# NOTE we use this to avoid divide by zero errors


@nb.jit(nb.float64[:, ::1](nb.int64), parallel=True, cache=True, nogil=True)
def basis(dim: int) -> NDArray[np.float64]:  # pragma: no cover # can't trace numba
    """Create a basis for rotating the simplex onto the origin of dim - 1."""
    res = np.empty((dim, dim))
    for i in nb.prange(0, dim - 1):
        num = i + 1
        frac = 1 / (num + 1)
        res[i, :num] = -((frac / num) ** 0.5)
        res[i, num] = (1 - frac) ** 0.5
        res[i, num + 1 :] = 0
    res[-1] = (1 / dim) ** 0.5
    return res


@dataclass(frozen=True, slots=True)
class Simplex(Topology):
    """The topology of the simplex.

    This represents all points in R^n s.t. 0 < x_i and Σx_i = 1.
    """

    def to_inf_ball(self, points: NDArray[np.float64]) -> NDArray[np.float64]:
        tiny = np.finfo(points.dtype).smallest_normal
        d = points.shape[-1]

        simp_direc = points - np.full(d, 1 / d)
        isimp_a = d * np.maximum(-simp_direc, simp_direc / (d - 1)).max(-1)

        cube_direc = simp_direc @ basis(d)[: d - 1].T
        icube_a = np.max(np.abs(cube_direc), -1) + tiny

        return np.clip(cube_direc * (isimp_a / icube_a)[..., None], -1, 1)  # type: ignore

    def from_inf_ball(self, points: NDArray[np.float64]) -> NDArray[np.float64]:
        tiny = np.finfo(points.dtype).smallest_normal
        d = points.shape[-1]

        icube_a = np.max(np.abs(points), -1)

        simp_direc = np.insert(points, d, 0, -1) @ basis(d + 1)
        isimp_a = (d + 1) * np.maximum(-simp_direc, simp_direc / d).max(-1) + tiny

        raw = simp_direc * (icube_a / isimp_a)[..., None] + np.full(d + 1, 1 / (d + 1))
        return np.clip(raw, 0, 1)  # type: ignore


@cache
def simplex() -> Simplex:
    """Create the topology of the simplex."""
    return Simplex()



"""Test for projection to plane."""

import numpy as np

import homeotopy


def test_known() -> None:
    """Test projection of known points."""
    plane = homeotopy.plane()

    plane_points = np.array(
        [[0, 0, 0], [np.inf, np.inf, -np.inf], [3, -5, 1 / 2]], "f8"
    )
    inf_ball_points = np.array([[0, 0, 0], [1, 1, -1], [3 / 4, -5 / 6, 1 / 3]], "f8")
    assert np.allclose(plane.to_inf_ball(plane_points), inf_ball_points)
    assert np.allclose(plane.from_inf_ball(inf_ball_points), plane_points)


def test_random() -> None:
    """Test projecting random points satisfy invariants."""
    rng = np.random.default_rng(0)
    plane = homeotopy.plane()

    inf_ball_points = rng.uniform(-1, 1, (3, 4, 5))
    plane_points = plane.from_inf_ball(inf_ball_points)
    assert np.allclose(plane.to_inf_ball(plane_points), inf_ball_points)



from dataclasses import dataclass
from functools import cache

import numpy as np
from numpy.typing import NDArray

from ._homeomorphism import Topology


@dataclass(frozen=True, slots=True)
class Sphere(Topology):
    """The topology of the unit sphere.

    This represents all points in R^n s.t. ||x||_2 = 1, except for the point
    (1, 0, ..., 0). That point is considered the boundary of the space, and will
    be mapped the largest closed point in some other topologies.

    Remarks
    -------
    If you had points on the surface of another p-ball, you could use the ball
    homeomorphism to first map them onto the surface of the 2-ball, and then
    apply this homeomorphism.
    """

    # NOTE for both of these we need to special case (1, 0, ..., 0) to (1, 1, ..., 1) and vice versa
    def to_inf_ball(self, points: NDArray[np.float64]) -> NDArray[np.float64]:
        tiny = np.finfo(points.dtype).smallest_normal
        scale = 1 - points[..., :1]

        normal = np.clip(np.tanh(points[..., 1:] / np.maximum(scale, tiny)), -1, 1)
        return np.where(scale <= 0, 1, normal)

    def from_inf_ball(self, points: NDArray[np.float64]) -> NDArray[np.float64]:
        big = np.finfo(points.dtype).max
        with np.errstate(divide="ignore"):
            plane = np.arctanh(points)

        s2 = (plane[..., None, :] @ plane[..., None])[..., 0]
        s2p = np.minimum(s2 + 1, big)

        x0 = np.where(s2 == np.inf, 1, (s2 - 1) / s2p)
        xns = np.where(s2 == np.inf, 0, 2 * plane / s2p)
        return np.concatenate([x0, xns], -1)


@cache
def sphere() -> Sphere:
    """Create a topology fot the unit sphere."""
    return Sphere()



from dataclasses import dataclass
from functools import cache

import numpy as np
from numpy.typing import NDArray

from ._homeomorphism import Topology


@dataclass(frozen=True, slots=True)
class Cube(Topology):
    """The topology of the unit hyper cube.

    This represents all points in R^n s.t 0 < x_i < 1, although the boundary
    should also work.
    """

    def to_inf_ball(self, points: NDArray[np.float64]) -> NDArray[np.float64]:
        return np.clip(points * 2 - 1, -1, 1)

    def from_inf_ball(self, points: NDArray[np.float64]) -> NDArray[np.float64]:
        return np.clip((points + 1) / 2, 0, 1)


@cache
def cube() -> Cube:
    """Create a topology of the unit cube."""
    return Cube()



from __future__ import annotations

from collections.abc import Callable, Iterable
from typing import Protocol

class _Decorator(Protocol):
    def __call__[**P, R](self, func: Callable[P, R]) -> Callable[P, R]: ...

class Type:
    pass

class Argument(Type):
    def __call__(self, *_: Argument) -> Type: ...

class Scalar(Argument):
    def __getitem__(self, val: slice | tuple[slice, ...]) -> Argument: ...

int64: Scalar
float64: Scalar

def jit(
    sig: Type, /, *, parallel: bool = ..., cache: bool = ..., nogil: bool = ...
) -> _Decorator: ...
def prange(start: int, stop: int | None = ...) -> Iterable[int]: ...
